From mathcomp Require Import ssreflect ssrfun ssrbool ssrnat eqtype seq.
From Kont Require Import prelude rels domains heaps hprop callcc.
Require Import Eqdep.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(******************************************************)
(* Lemmas for discharging HTTcc proof obligations     *)
(******************************************************)

Open Scope ccscope.

Section PullDef.
Variables (A : Type) (i : heap).
Variables (q : A -> heap -> Prop).
Variables (s: spec A).

Lemma vrfD : (def i -> CCC.verify i s q) -> CCC.verify i s q.
Proof. by move => H Di; apply: (H Di). Qed.
End PullDef.

Section EvalDo.
Variables (A B : Type) (i : heap).
Variables (q : A -> heap -> Prop) (r : B -> heap -> Prop).
Variables (s1: spec A) (s2 : A -> spec B).

Lemma val_do :
         s1.1 i -> 
         (forall x k, def k -> s1.2 x i k -> q x k) ->
         CCC.verify i s1 q.
Proof. by move => H H1 ; split => // x k Dk H2; apply: (H1 x k H2 Dk). Qed.

Lemma bnd_do :   
        s1.1 i -> 
        (forall x k, s1.2 x i k -> CCC.verify k (s2 x) r) -> 
        CCC.verify i (CCC.bind_s s1 s2) r.
Proof.
move=>H1 H2 D; split=>[|y m].
- by split=>[//|y m] Dm H3; case: (H2 _ _ H3 Dm).
by case=>x[h][Dh][H3] H4 Dm; case: (H2 _ _ H3 Dh)=>H5; apply.
Qed.

Lemma bnd_val : 
        CCC.verify i s1 (fun y m => CCC.verify m (s2 y) r) <->
        CCC.verify i (CCC.bind_s s1 s2) r.
Proof.
split.
+ move=>H; apply:vrfD;case/H=>H1 H2; apply: bnd_do=>//.
 by move=> x m H3; apply: vrfD; move/(@H2 x m H3).
move=>H; apply:vrfD;case/H=>[[H1 H2]] H3; split=>//{H0}.
move=> y m H4 Dm; split=>//; first by apply: H2.
move=>{H0} z k H5 Dk; apply: H3=>//. 
by exists y, m.
Qed.

End EvalDo.

Section EvalRet.

Variables (A B : Type) (i : heap).
Variables (q : A -> heap -> Prop) (r : B -> heap -> Prop).
Variables (s1: spec A) (s2 : A -> spec B).

Lemma val_ret v: 
       q v i -> CCC.verify i (CCC.ret_s v) q. 
Proof. by move=>H; apply val_do => // x k Dk [->]->. Qed.

Lemma bnd_ret v: 
        CCC.verify i (s2 v) r -> CCC.verify i (CCC.bind_s (CCC.ret_s v) s2) r. 
Proof. by move=>H; apply: bnd_do => // x h [->]->. Qed.
End EvalRet.

Section EvalBind.

Variables (A B C: Type) (i : heap).
Variables (q : A -> heap -> Prop) (r : C -> heap -> Prop).
Variables (s1: spec A) (s2 : A -> spec B) (s3: B -> spec C).

Lemma bnd_bnd: CCC.verify i (CCC.bind_s s1 (fun x => CCC.bind_s (s2 x) s3)) r -> 
                 CCC.verify i (CCC.bind_s (CCC.bind_s s1 s2) s3) r. 
Proof.
move=>H ; apply: vrfD => Di. 
case: H => // [H1] H2 ; case: H1 => P1 H3.
apply: bnd_do => // /= [ | y k [x] [h] [Dh] [Q1] Q2].
+ by split => // x m Dm Qm; move: (H3 x m Dm Qm)=> [H4] _.
move => Dk; move: (H3 x h Dh Q1) => [_ H6].
split => [| z m Q3 Dm]; first by apply: (H6 y k Dk Q2).
apply: (H2 z m) => // /=.
by exists x; exists h; repeat split => //; exists y; exists k.
Qed.
End EvalBind.


Section EvalRead.

Variables (A B: Type) (v: A) (x : ptr).
Variables (i h : heap) (q: A -> heap -> Prop) (r : B -> heap -> Prop).
Variable (s : A -> spec B).

Lemma val_read : 
        (def (x:-> v :+ h) -> q v (x :-> v :+ h)) -> 
           CCC.verify (x :-> v :+ h) (CCC.read_s A x) q.
Proof.
move=> H; apply val_do => //=; first by exists v;exists h.
by move=>y k Dk [E H1]; rewrite E in  Dk *; rewrite (H1 v h)//; apply: (H Dk).
Qed.

Lemma bnd_read: 
        CCC.verify (x :-> v :+ i) (s v) r -> 
        CCC.verify (x :-> v :+ i) (CCC.bind_s (CCC.read_s A x) s) r.
Proof.
move => H; apply: bnd_do => //= [|y k [-> H2]] ; first by exists v ; exists i.
by move: (H2 v i (erefl (x:-> v :+ i))) ->.
Qed.
End EvalRead.

Section EvalWrite.
Variables (A B C : Type) (v : A) (w : C) (x : ptr).
Variables (i h : heap) (q: unit -> heap -> Prop) (r : B -> heap -> Prop).
Variable (s : unit -> spec B).

Lemma val_write: 
        (def (x :-> v :+ i) -> q tt (x :-> v :+ i)) -> 
        CCC.verify (x :-> w :+ i) (CCC.write_s x v) q.
Proof.
move => H; apply val_do => //=; first by exists C; exists w ;exists i .
move => u k Dk H1.
move: (H1 C w i (erefl (x:-> w :+ i))) => [E]->; rewrite E in Dk *.
by apply: (H Dk).
Qed.

Lemma bnd_write: 
        CCC.verify (x :-> v :+ i) (s tt) r -> 
        CCC.verify (x :-> w :+ i) (CCC.bind_s (CCC.write_s x v) s) r.
Proof.
move => H; apply: bnd_do =>//= [|u k H1]; first by exists C; exists w; exists i.
by move: (H1 C w i (erefl (x:-> w :+ i)))=> [->]->.
Qed.
End EvalWrite.

Section EvalAlloc.
Variables (A B: Type) (v: A) (x : ptr).
Variables (i h : heap) (q: ptr -> heap -> Prop) (r : B -> heap -> Prop).
Variables (s : ptr -> spec B).

Lemma val_alloc : 
        (forall x, def (x :-> v :+ i) -> q x (x :-> v :+ i)) -> 
        CCC.verify i (CCC.alloc_s v) q.
Proof.
move=> H; apply: val_do=>//.
by move=> //= y k Dk H0 ; rewrite H0 in Dk *; apply: H.
Qed.

Lemma bnd_alloc : 
        (forall x, CCC.verify (x :-> v :+ i) (s x) r) ->
        CCC.verify i (CCC.bind_s (CCC.alloc_s v) s) r.
Proof. by move => H ; apply: bnd_do => //= y m ->. Qed.
End EvalAlloc.

Section EvalDealloc.
Variables (A B : Type) (x: ptr) (v: A).
Variables (i : heap) (q: B -> heap -> Prop) (r : unit -> heap -> Prop).
Variables (s : unit -> spec B).

Lemma val_dealloc: 
        (def i -> r tt i) -> 
        CCC.verify (x :-> v :+ i) (CCC.dealloc_s x) r.
Proof.
move=> H; apply: vrfD => H0.
apply: val_do => //= ; first by [exists A; exists v; exists i].
by move=>y m Dm H1; rewrite H1 freePtUn in Dm *; last by []; elim y; apply: H.
Qed.

Lemma bnd_dealloc: 
        CCC.verify i (s tt) q -> 
        CCC.verify (x :-> v :+ i) (CCC.bind_s (CCC.dealloc_s x) s) q.
Proof. 
move=> H ; apply: vrfD => H0.
apply: bnd_do => //= ; first by [exists A; exists v ; exists i].
by move => y k ->; rewrite freePtUn; elim y.
Qed.
End EvalDealloc.

Section eval_CCCR.
Variables (A: Type) (ssf  sse : heap -> spec A) (R: heap -> heap -> Prop).
Variables (i : heap) (rr : A -> heap -> Prop). 

Lemma val_cccR :
          (CCC.cccR_s R ssf sse).1 i ->
              CCC.verify i (sse i) rr ->
               (forall x m, def m -> (CCC.postcomp R (ssf i).2) x i m -> rr x m ) ->
                  CCC.verify i (CCC.cccR_s R ssf sse) rr.
Proof.
move =>/= [Pe H0] H1 H2; apply: vrfD=> D; move: (H1 D)=>/= [_ H3].
apply: val_do=>//= x k Dk [|[s][Ris][H4 H5]]; first by move/(H3 x k); move/(_ Dk).
by apply: (H2 x k Dk); exists s.
Qed.
End eval_CCCR.

Section CCC_Algebraic.
Variables (A B: Type) (sf se : heap -> spec A) (R: heap -> heap -> Prop).
Variables (i : heap) (q: B -> heap -> Prop) (s2 : A -> spec B). 

Lemma ccc_prealg j : 
    CCC.bind_pre (CCC.cccR_s R sf se) s2 j <-> 
      (CCC.cccR_s R (fun k => CCC.bind_s (sf k) s2)  
        (fun k => CCC.bind_s (se k) s2)).1 j.
Proof.
split=>/=.
+ move =>/= [/=[H0] H1] H2.
 split; first by split=>// x m Dm H3; apply: (H2 x m Dm); left.
 move => s Ds H3; split=>[| x m Dm H4] ; first by apply: (H1 s Ds H3).
 by apply: (H2 x m Dm); right; exists s.
move=>/= [[H0] H1] H2.
try repeat split=>//=;first by move => k Dk Rik; move:(H2 k Dk Rik)=> [H4]_.
move=> x m Dm [H3 | H3]; first by apply: H1.
by case H3=> s [Ds] [Ris H4]; move: ( H2 s Ds Ris)=>[_]H5; apply: H5.
Qed.

Lemma ccc_postalg  y k m :
   CCC.bind_post (CCC.cccR_s R sf se) s2 y k m <-> 
      (CCC.cccR_s R (fun j => CCC.bind_s (sf j) s2)
        (fun j => CCC.bind_s (se j) s2)).2 y k m.
Proof.
split=>/=.
+move=>[x][s][Ds][H0 H1]; case: H0=> H0; first by left; exists x; exists s.
 right; case: H0=>[j][Dj][Rij H2]; exists j; try repeat split=>//.
 by exists x; exists s.
move => [ [x][s][Ds][H0 H1] | [s][Ds][Ris][x][j][Dj][H0 H2]]. 
+ by exists x; exists s; try repeat split=>//; left.
by exists x; exists j; try repeat split=>//; right; exists s.
Qed.

Lemma callccR_algebraic :
  CCC.verify i (CCC.cccR_s R (fun j => CCC.bind_s (sf j) s2) 
                        (fun j => (CCC.bind_s (se j) s2))) q <->
    CCC.verify i (CCC.bind_s (CCC.cccR_s R sf se) s2) q.
Proof.
split.
+move => H; apply: vrfD=>D; case: (H D)=>/= [[[H0 H1] H2] H3].
 apply: bnd_do=>/=; first by split=>//= h Dh Rih; move: ( H2 h Dh Rih)=>[H4]_.
 move => x k [H4 | H4] Dk.
 -split=>/=[| y m H5 Dm]; first by apply: (H1 x k Dk H4).
  by apply: (H3 y m)=>//=; left; exists x; exists k; repeat split.
 split=>/=[| y m H5 Dm]; case: H4=>s[Ds][Ris]H6.
 - by move:(H2 s Ds Ris)=>[_ H7]; apply: H7.
 apply: (H3 y m)=>//; right; exists s; repeat split=>//.
 by exists x; exists k; repeat split.
move => H; apply: vrfD=>D; case: (H D)=>/= [[[H0 H1] H2] H3].
apply: val_do=>/= [| y m Dm H6].
+split=>[| k Dk  Rik]; first by split=>// x m Dm H4; apply: (H2 x m Dm)=>/=;left.
 move:(H1 k Dk Rik)=> H4; split=>// y s Ds H5; apply: (H2 y s Ds)=>/=; right.
 by exists k.
apply: (H3 y m _ Dm); case: H6=> [[x][s][Ds][H4 H5]|[s][Ds][Ris]H7].
+ by exists x; exists s; repeat split=>//=; left.
case: H7 => [x][h][Dh][Rih]H7; exists x; exists h; repeat split=>//.
by right; exists s; split.
Qed.

Lemma bnd_cccR:
        CCC.verify i (CCC.bind_s (se i) s2) q ->
           (forall j , R i j -> CCC.verify j (CCC.bind_s (sf i) s2 )  q ) ->
               CCC.verify i (CCC.bind_s (CCC.cccR_s R sf se) s2) q.
Proof.
move => H1 H2; apply: vrfD=> D; rewrite -callccR_algebraic.
move: (H1 D)=>/= [[H0 H3] H4].
apply: (val_cccR)=>//.
+ rewrite -ccc_prealg; repeat split=>//=.
  - by move=>/= s Ris Ds;move:(H2 s Ris); move/(_ Ds)=>[[H5] _] _.
  move => x m Dm H5.
  case: H5=>[|[s][Ris][Ds H5]]; first by move/(H3 x m Dm). 
  by move:(H2 s Ris); move/(_ Ds)=>[[_ H7] _]; apply: H7.
move=> x m Dm [s][Ds][Ris] [y][h][Dh][H6 H7].
move:(H2 s Ris); move/(_ Ds)=>[_]H8; apply: (H8 x m _ Dm)=>/=.
by exists y; exists h.
Qed.

(* Proof of BindCC without algebraicity*)
(* Lemma bnd_cccR: *)
(*         CCC.verify i (CCC.bind_s (se i) s2) q -> *)
(*            (forall j , R i j -> CCC.verify j (CCC.bind_s (sf i) s2 )  q ) -> *)
(*                CCC.verify i (CCC.bind_s (CCC.cccR_s R sf se) s2) q. *)
(* Proof. *)
(* move=> H0 H; apply: vrfD; move/H0=>{H0} [[H0 H1] H2] D;split=>/=.  *)
(* + split=>//=; first by split=>//= s; move/H=> H3; move/H3=>{H3} [[H3 _] _]. *)
(*  move=> y m Dm [H3 |[s][Ris][Ds]H4] ; first by apply: (H1 y m Dm H3). *)
(*  by move: (H s Ris Ds)=>[[_ H6] _]; apply: (H6 y m Dm H4). *)
(* move=> y m [x][s][Ds]/=[[H3| ]]. *)
(* + by move=> H4; apply: (H2 y m); exists x; exists s. *)
(* move=>[h][Rih][Dh]H3 H4. *)
(* by move: (H h Rih Dh)=>[_]; apply; exists x; exists s. *)
(* Qed. *)
End CCC_Algebraic.

(* Specialised lemmata for examples when pf _ = xPredT *)
Section eval_CCC.
Variables (A B: Type) (se : heap -> spec A) (R: heap -> heap -> Prop).
Variables (i : heap) (qf: heap -> post A) (s2 : A -> spec B). 
Variables (r : A -> heap -> Prop)  (q : B -> heap -> Prop).

Lemma val_ccc :
  CCC.verify i (se i) r ->
     (forall x m, def m -> 
          (CCC.postcomp R (qf i)) x i m -> r x m ) ->
          CCC.verify i (CCC.ccc_T R qf se) r.
Proof.
move =>/= H1 H2; apply: vrfD; case/H1=>/= H3 H4.
apply: val_do=>//= x k Dk [|[s][Ris][H5 H6]]; 
  first by move/(H4 x k); move/(_ Dk).
by apply: (H2 x k Dk); exists s.
Qed.

Lemma bndccSP :
 CCC.verify i (CCC.bind_s (CCC.cccR_s R (fun j => (xPredT, qf j)) se) s2) q ->
     CCC.verify i (CCC.bind_s (CCC.ccc_T R qf se) s2) q.
Proof.
by move=> H; apply: vrfD; case/H=>/= [/= [[P0 _] P1]] H0; do 2 split=>//.
Qed.

Lemma bnd_ccc:
     CCC.verify i (CCC.bind_s (se i) s2) q ->
      (forall x m , (CCC.postcomp R (qf i)) x i m -> CCC.verify m (s2 x) q) ->
             CCC.verify i (CCC.bind_s (CCC.ccc_T R qf se) s2) q.
Proof.
move=> H0 H1; apply: bnd_cccR=>// j  Rij; apply: vrfD=>D.
by apply: bnd_do=>//= x m H2; apply: vrfD; case/( H1 x m)=>//=; by exists j.
Qed.

Lemma bnd_ccc':
   CCC.verify i (CCC.bind_s (se i) s2) q ->
      (forall x m , (CCC.postcomp R (qf i)) x i m -> CCC.verify m (s2 x) q) ->
               CCC.verify i (CCC.bind_s (CCC.ccc_T R qf se) s2) q.
Proof.
move=>H1 H2; apply: vrfD=>Di; apply: bnd_do=>/=.
- by split=>//; case: H1=>//; case. 
move=>x k [{H2} H2|[s [H3 [Ds H4]]]]; last by apply: H2; exists s.
case: H1=>//= [[H3 H4] H5] Dk. 
split=>[|y m H6]; first by apply: H4 Dk H2.
by apply: H5; exists x, k.  
Qed.
End eval_CCC.
 
Section EvalAbort.
Variables (X Y: Type) (sg st : spec Y) (s2: Y -> spec X).
Variables (i : heap) (r: X -> heap -> Prop ).

Lemma val_abort:
       (CCC.abort_s X sg st).1 i ->  CCC.verify i (CCC.abort_s X sg st) r.
Proof. by  []. Qed.
End EvalAbort.

(* The following lemma implements the algebraic property for abort: *)
(* abort z >>= e =  abort z  *)
(* in the SK idiom *)

Section Abort_Algebraic.
Variables (A B: Type) (sg st : spec A) (s2: A -> spec B).
Variables (i : heap) (q : B ->  heap -> Prop) .

Lemma abort_algebraic :
  CCC.verify i (CCC.abort_s B sg st) q <->
    CCC.verify i (CCC.bind_s (CCC.abort_s A sg st) s2) q.
Proof.
split; move=> H Di /=; case: H=> //= P0 H1; last by case:P0.
split=> [| y m [w][t][Dt][F H] Dm]; first by split=>// x m Dm F.
by apply: False_ind.
Qed.

Lemma bnd_abort:
       (CCC.abort_s B sg st).1 i -> 
         CCC.verify i (CCC.bind_s (CCC.abort_s A sg st) s2) q.
Proof. by rewrite -abort_algebraic; apply: val_abort. Qed.
End Abort_Algebraic.

(* abort g  ( e_1; e_2 ) = abort ( g . ( >\ t . t ; e_2)) e_1 *)

Section Eval_throw.
Variables (X Y: Type) (sg: spec Y) (e : Y). 
Variables (i : heap) (r: X -> heap -> Prop ).

Lemma val_throw:
     sg.1 i ->
     sg.2 e i i ->  CCC.verify i (CCC.throw_s X sg e) r.
Proof. 
move=> H H2; apply: val_abort=>//=; split=>//= i1 [H1] ->.
by apply:val_ret=>//=.
Qed.
End Eval_throw.

Section Throw_Algebraic.
Variables (A B: Type) (sg : spec A) (s2: A -> spec B).
Variables (i : heap) (X: A) (q : B ->  heap -> Prop) .

Lemma throw_algebraic :
  CCC.verify i (CCC.throw_s B sg X) q <->
    CCC.verify i (CCC.bind_s (CCC.throw_s A sg X) s2) q.
Proof.
split;move/abort_algebraic; [rewrite -abort_algebraic=> H1|];  
by rewrite (@abort_algebraic A B _ _ s2).
Qed.

Lemma bnd_throw:
       (CCC.throw_s B sg X).1 i -> 
         CCC.verify i (CCC.bind_s (CCC.throw_s A sg X) s2) q.
Proof. by rewrite -throw_algebraic.  Qed.
End Throw_Algebraic.


(******************************************************************************)
(*                    Frame Rule for SK                                       *) 
(******************************************************************************)


Definition lolli (p : _ -> Prop) q i m := 
  forall i1 h, i = i1 :+ h -> def i -> p i1 -> 
    exists m1, m = m1 :+ h /\ def m /\ q i1 m1.

Notation "p '--o' q"   := (lolli p q) (at level 75) : ccscope.

Lemma antiframe p q i m h : 
        def (i :+ h) -> (p --o q) (i :+ h) (m :+ h) -> (p --o q) i m.

Proof.
move=>D1 H i2 m2 E D2 H1; rewrite {i}E in H D1 D2 *.
move: (H i2 (m2 :+ h)); rewrite unA; case/(_ (erefl _) D1 H1)=>h2 [E][D3].
rewrite unA in E; exists h2; rewrite (unKhl D3 E).
by rewrite E in D3; rewrite (defUnl D3).
Qed.

(* p --o q is local *)
Lemma locality p q i1 m h : 
        def (i1 :+ h) -> (p # top) i1 -> (p --o q) (i1 :+ h) m -> 
          exists m1, m = m1 :+ h /\ def m /\ (p --o q) i1 m1.
Proof.
move=>D [h1][h2][E][H1] _ H2; rewrite {i1}E in D H2 *.
move: (H2 h1 (h2 :+ h)); rewrite unA; case/(_ (erefl _) D H1)=>m1 [E][D2].
rewrite {m}E unA in H2 D2 *; exists (m1 :+ h2); do !split=>//.
by apply: antiframe D H2.
Qed.

Lemma fr_pre p i j : (p # top) i -> (p # top) (i :+ j).
Proof. 
by case=>i1 [i2][->][H] _; rewrite -unA; exists i1; exists (i2 :+ j). 
Qed.

Definition fr A (s : spec A) : spec A := 
  (s.1 # top, fun x => s.1 --o s.2 x).

Prenex Implicits fr.

Notation "[ s ]" := (fr s).

Section SepFrame.
Variables (A : Type) (s : spec A). 

Lemma frame i j (r : A -> heap -> Prop) : 
        CCC.verify i [s] (fun y m => def (m :+ j) -> r y (m :+ j)) ->
        CCC.verify (i :+ j) [s] r.
Proof.
move=>H D; case: (H (defUnl D))=>H1 H2.
split=>[|y m]; first by apply: fr_pre.
case/(locality D H1)=>m1 [->][D1]; move/H2.
by apply; apply: defUnl D1.
Qed.

Lemma frame0 i r : CCC.verify i s r -> CCC.verify i [s] r.
Proof.
move=>H D; case: (H D)=>H1 H2.
split=>[|y m]; first by exists i; exists empty; rewrite unh0.
move/(_ i empty); rewrite unh0; case/(_ (erefl _) D H1)=>m1.
by rewrite unh0=>[[<-]][_]; apply: H2.
Qed.

Lemma frame1 i (r : A -> heap -> Prop) :
        CCC.verify empty s (fun y m => def (m :+ i) -> r y (m :+ i)) -> 
        CCC.verify i [s] r.
Proof. by move=>H; rewrite -[i]un0h; apply: frame; apply: frame0. Qed.
End SepFrame.

(******************************************************************************)
(*                   Lemmas for pulling out ghost variables                   *)
(******************************************************************************)

Section Ghosts.
Variables (A : Type) (s : spec A) (p : pre).

Lemma allC (B1 B2 : Type) (q : B1 -> B2 -> post A) :
       CCC.conseq s p (fun y i m => forall x1 x2, q x1 x2 y i m) <->
       CCC.conseq s p (fun y i m => forall x, q x.1 x.2 y i m).
Proof.
split=>//= H1 i H2 Di; case: (H1 i H2 Di) => S1 H3; split => //= y m S2 Dm.
- by move => [x1 x2]; apply: H3.
by move => x1 x2; apply: (H3 y m S2 Dm (x1,x2)).
Qed. 
   
Lemma impC (B : Type) (q1 q2 : heap -> B -> Prop) (r : B -> post A) :
        CCC.conseq s p (fun y i m => forall x, q1 i x -> q2 i x -> r x y i m) <->
        CCC.conseq s p (fun y i m => forall x, q1 i x /\ q2 i x -> r x y i m).
Proof.
split => //= H1 i H2 Di; case: (H1 i H2 Di) => S1 H3; split => //= y m S2 Dm x.
- by case => H4 H5 ; apply: (H3 y m S2 Dm).
by move => H4 H5 ; apply: (H3 y m S2 Dm).
Qed.

Lemma ghE (B : Type) (q : heap -> B -> Prop) (r : B -> post A) : 
        (forall i, p i -> def i -> exists x, q i x) ->
        (forall i x, q i x -> p i -> def i ->
           CCC.verify i s (fun y m => r x y i m)) ->
        CCC.conseq s p (fun y i m => forall x, q i x -> r x y i m).
Proof.
move=>H1 H2 i H3 Di; case: (H1 i H3 Di)=>x H4.
case: (H2 i x H4 H3 Di Di)=>H5 _; split=>// y m H6 Dm z H7.
by case: (H2 i z H7 H3 Di Di)=>_; apply.
Qed.

End Ghosts.
Definition gh := (allC, impC).