(***************************************************)
(*    HTTcc Examples for the ICFP'13 submbission   *)
(*          Contrasting Abort and Throw            *)
(***************************************************)

From mathcomp Require Import ssreflect ssrfun ssrbool ssrnat eqtype seq div.
From Kont Require Import rels prelude heaps domains hprop callcc srules.
Require Import Eqdep.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Open Scope ccscope.

Parameter (X : lattice).

(* In this file we illustrate the semantic differences between between
throwing and aborting, by means of the following example. *)

Program 
Definition incr ( x : ptr) : 
  SK X ( fun i => exists h, exists v: nat , 
                i= x :-> v :+ h , 
       fun y i m => forall h v, 
         i = x :-> v :+ h -> m = x :-> v .+1 :+ h /\ y = tt) := 
  Do ( z <-- ! x ; x ::= z .+1 ).
Next Obligation.     
move=> i [h][v]->; apply: bnd_read; apply: vrfD=>D; apply:val_write=> _ h1 v1.
by move/(cancel D)=>[->]_->; split=>//.
Qed.

Program
Definition swap T (x y : ptr) :
  SK X ( fun i =>  exists h, exists v: T, exists w: T, 
                i= x :-> v :+ y :-> w :+ h , 
       fun _ i m => forall  h (v w : T), 
         i = x :-> v :+ y :-> w :+ h -> 
                   m = x :-> w :+ y :-> v :+h) := 
  Do ( p <-- @CCC.read X T x ;  q <-- @CCC.read X T y; x ::= q ;; y ::= p).
Next Obligation.
rewrite !gh; apply: ghE; first by move=> i [h][v][w]-> _; by exists (h,v,w).
move=> i [[h]]v w /= -> [h1][v1][w1] E; rewrite -unA=>D.
do 2 (apply: bnd_read; rewrite unCA); apply: bnd_write; rewrite unCA.
by apply val_write; rewrite unCA !unA=> _.
Qed.
Arguments swap {T} x y.

Section Throwswap.
Variables (x y : ptr) (V :nat).


Definition tswapP j  : pre :=
     fun i => forall h (v w : nat), 
                 j = x :-> v :+ (y :-> w :+ h) -> 
                    i = x :-> v.+1  :+ ( y :-> w :+ h). 

Definition tswapQ1: post unit := fun y i m => False.
Definition tswapQ2 : post unit := 
  fun r i m => forall h (v w : nat), 
                 i = x :-> v :+ y :-> w :+ h -> 
                    m = x :-> w :+  y :-> v :+ h.

Program
Definition throwswap : SK X
  ( fun i =>  exists h, i = x :-> V :+ y :-> V :+ h,
     fun r i m =>  forall h, i = x :-> V :+  y :-> V :+ h ->
                       m = x :-> V :+ y :-> V.+2 :+ h /\ r = tt) :=
  Do (c <-- @CCC.callcc X (SigSK X)  
(* Rf *)
                (fun j i => forall h,
                              j = x :-> V :+ y :-> V  :+ h ->
                              i = x :-> V.+1 :+ y :-> V  :+ h) 
(* Q_f *)
                ( fun j r i m => 
                                m = i  /\
                               CCC.conseq (spec_of r) (tswapP m) tswapQ2)
(* j -> se *)
                (fun j =>  
                     (fun i =>
                       exists h, i = x :-> V :+ y :-> V :+ h /\ j = i
                       , fun z i m => 
                              m = i /\
                             CCC.conseq (spec_of z) 
                             (tswapP j) tswapQ1))
        (fun j ex1 =>
           Do (ret (BigSK (throw ex1 (BigSK (@swap nat x y ;; ret tt))))));
      incr x ;;
      comm c).
Next Obligation.
move =>i[h][->];rewrite -unA=>J; apply: vrfD=>D; apply: val_ret;split=>//= k.
move/(_ h V V J )=>->; apply: val_throw=>//=.
+ by split=>// h1; rewrite J -!unA; move/(cancel D)=>[_]D1; 
  move/(cancel D1)=>[_]_<-.
split=>// {k} k1; move/(_ h V.+1 V (erefl _))=>->; apply: bnd_do=>//=.
+ by exists h; exists V.+2; exists V; rewrite unA.
move=> tt k ; rewrite unA; move/(_ h V.+2 V (erefl _))=>->.
apply: val_ret=> h1 v1 w1; rewrite -!unA.
have D1: def (x :-> V.+2 :+ ( y :-> V :+ h)) by rewrite !defPtUn in D *.
by case/(cancel D1)=>->D2; case/(cancel D2)=><- _ <-.
Qed.
Next Obligation.
move=> i[h]; rewrite -unA=>->; apply: vrfD=>D; apply: bnd_ccc.
+ apply: bnd_do; first by exists h; rewrite !unA.
 move=>/= c k [->] H; apply: bnd_do=>/=; first by hhauto.
 move=>tt {k} k; case/(_ (y :-> V :+ h) V (erefl _))=>-><-; apply: vrfD=>D1.
 have P: tswapP (x :-> V :+ (y :-> V :+ h)) (x :-> V.+1 :+ (y :-> V :+ h))
   by move=> h1 v1 w1; case/(cancel D)=>-> D2; case/(cancel D2)=>-> _ ->.
 by case: (H _ P D1)=>[P0 H1]; apply: val_do=>// [tt1] {k} k Dk;
   move/H1/(_ Dk); apply: False_ind.
move=> c k [k1][_]; case ; move/(_ h); rewrite -!unA; move/(_ (erefl _))=>->.
case=>-> H; apply: vrfD=>D1; apply: bnd_do=>//=; first (by hhauto). 
move=> {k} tt k; move/(_ ( y :-> V :+ h) V.+1 (erefl _))=>[->]_.
have P: tswapP (x :-> V.+1 :+ (y :-> V :+ h)) (x :-> V.+2 :+ (y :-> V :+ h))
   by move=> h1 v1 w1; move/(cancel D1)=>[->]D2; case/(cancel D2)=>-> _ ->.
apply: vrfD; case/( H _ P)=>[P1 H1]; apply: val_do=>//= tt1{k}k Dk; case: tt1.
move/H1/(_ Dk)/(_ h V.+2 V); rewrite -!unA;move/(_ (erefl _))=>-> h1.
by rewrite -!unA; case/(cancel D)=> _ D2; case/(cancel D2)=> _ _ <-. 
Qed.

End Throwswap.

(* If we abort with swap instead of throwing it back to the *)
(* continuation, we get different results after running the continuation *)
(* twice: *)

Section Abortswap.
Variables (x y : ptr) (V :nat).


Definition aswapP j  : pre :=
     fun i => forall h (v w : nat), 
                 j = x :-> v :+ (y :-> w :+ h) -> 
                    i = x :-> v.+1  :+ ( y :-> w :+ h). 

Definition aswapQ1: post unit := fun y i m => False.
Definition aswapQ2 : post unit := (CCC.ret_s tt).2.

Program
Definition abortswap : SK _
  ( fun i =>  exists h, i = x :-> V :+ y :-> V :+ h,
     fun r i m =>  forall h, i = x :-> V :+  y :-> V :+ h ->
                       m = x :-> V.+1 :+ y :-> V.+1 :+ h /\ r = tt) :=
  Do (c <-- @CCC.callcc X (SigSK X) 
(* Rf *)
                (fun j i => forall h,
                              j = x :-> V :+ y :-> V  :+ h ->
                              i = x :-> V.+1 :+ y :-> V  :+ h) 
(* Q_f *)
                ( fun j r i m => forall h,
                              j = x :-> V :+ y :-> V  :+ h ->
                              i = x :-> V.+1 :+ y :-> V :+ h ->
                                m = x :-> V :+ y :-> V.+1 :+ h  /\
                               CCC.conseq (spec_of r) (aswapP m) aswapQ2)
(* j -> se *)
                (fun j =>  
                     (fun i =>
                       exists h, i = x :-> V :+ y :-> V :+ h /\ j = i
                       , fun z i m => 
                              m = i /\
                             CCC.conseq (spec_of z) 
                             (aswapP j) aswapQ1))
        (fun j ex1 =>
         Do (ret (BigSK (abort _ ex1 (@swap nat x y;; ret (BigSK (ret tt)))))));
      incr x ;;
      comm c).
Next Obligation.
move=> i[h]; rewrite -unA=>[[->]]J; apply: vrfD=>D; apply: val_ret; split=>// k.
move/(_ h V V J)=>->;  apply: val_abort=>/=; split=>//=.
+ by split=>// h1; rewrite J -!unA; case/(cancel D)=> _ _ ->.
move=>/= {k} k [_ ->]; apply: bnd_do=>//=;
  first by exists h; exists V.+1; exists V; rewrite unA.
move=> _ {k} k; move/(_ h V.+1 V); rewrite -!unA; move/(_ (erefl _))=>->.
apply:val_ret=>//= h1; rewrite J -!unA; case/(cancel D)=> _ D1; case/(cancel D1).
by move=> _ _ <- _; split=>//; apply: val_ret.
Qed.
Next Obligation.
move=> i[h]; rewrite -unA=>->; apply: vrfD=>D; apply: bnd_ccc.
+ apply: bnd_do; first by exists h; rewrite !unA.
 move=>/= c k [->] H; apply: bnd_do=>/=; first by hhauto.
 move=>tt {k} k; case/(_ (y :-> V :+ h) V (erefl _))=>-><-; apply: vrfD=>D1.
 have P: aswapP (x :-> V :+ (y :-> V :+ h)) (x :-> V.+1 :+ (y :-> V :+ h))
   by move=> h1 v1 w1; case/(cancel D)=>-> D2; case/(cancel D2)=>-> _ ->.
 by case: (H _ P D1)=>[P0 H1]; apply: val_do=>// [tt1] {k} k Dk;
   move/H1/(_ Dk); apply: False_ind.
move=>c k [k1][_]; case ; move/(_ h); rewrite -!unA; move/(_ (erefl _))=>->.
move/(_ h);rewrite -!unA; move/(_ (erefl _) (erefl _))=>[->]H.
apply: vrfD=>D1; apply: bnd_do=>//=; first (by hhauto). 
case=> {k k1} k; move/(_ ( y :-> V.+1 :+ h) V (erefl _))=>[->]_.
have P: aswapP (x :-> V :+ (y :-> V.+1 :+ h)) (x :-> V.+1 :+ (y :-> V.+1 :+ h))
   by move=> h1 v1 w1; move/(cancel D1)=>[->]D2; case/(cancel D2)=>-> _ ->.
apply: vrfD; case/( H _ P)=>[P1 H1]; apply: val_do=>//= tt1{k}k Dk; case: tt1.
move/H1/(_ Dk)=>/= [_]-> h1; rewrite -!unA.
by  case/(cancel D)=> _ D2; case/(cancel D2)=> _ _ <-. 
Qed.


End Abortswap.

(* This mimics abort f e  by doing e ; throw f *)

Section MimicAbort.
Variables (x y : ptr) (V :nat).


Definition mswapP1 j  : pre :=
     fun i => forall h (v w : nat), 
                 j = x :-> v :+ (y :-> w :+ h) -> 
                    i = x :-> v.+1  :+ ( y :-> w :+ h). 



Definition mswapQ1: post unit := fun y i m => False.
Definition mswapQ2 : post unit := (CCC.ret_s tt).2.

Program
Definition mimicwswap : SK _
  ( fun i =>  exists h, i = x :-> V :+ y :-> V :+ h,
     fun r i m =>  forall h, i = x :-> V :+  y :-> V :+ h ->
                       m = x :-> V.+1 :+ y :-> V.+1 :+ h /\ r = tt) :=
  Do (c <-- @CCC.callcc X (SigSK X) 
(* Rf *)
                (fun j i => forall h,
                              j = x :-> V :+ y :-> V  :+ h ->
                              i = x :-> V :+ y :-> V .+1 :+ h) 
(* Q_f *)
                ( fun j r i m => 
                                m = i  /\
                               CCC.conseq (spec_of r) (mswapP1 m) mswapQ2)
(* j -> se *)
                (fun j =>  
                     (fun i =>
                       exists h, i = x :-> V :+ y :-> V :+ h /\ j = i
                       , fun z i m => 
                              m = i /\
                             CCC.conseq (spec_of z) 
                             (mswapP1 j) mswapQ1))
        (fun j ex1 =>
           Do (ret (BigSK (@swap nat x y ;; throw ex1 (BigSK (ret tt))))));
      incr x ;;
      comm c).
Next Obligation.
move=>/= i[h][->];rewrite -unA=>J; apply:vrfD=>D; apply: val_ret;split=>// k.
move/(_ h V V J)=>->; apply: bnd_do=>/=; first by (rewrite unA; hhauto).
move=>/= _ {k} k; move/(_ h V.+1 V); rewrite -!unA; move/( _ (erefl _))=>->.
apply: val_throw=>//=; split=>// h0; rewrite -!unA J; case/(cancel D)=> _ D1.
by case/(cancel D1)=> _ _ ->.
Qed.
Next Obligation.
move=> i[h]; rewrite -unA=>->; apply: vrfD=>D; apply: bnd_ccc.
+ apply: bnd_do; first by exists h; rewrite !unA.
 move=>/= c k [->] H; apply: bnd_do=>/=; first by hhauto.
 move=>tt {k} k; case/(_ (y :-> V :+ h) V (erefl _))=>-><-; apply: vrfD=>D1.
 have P: mswapP1 (x :-> V :+ (y :-> V :+ h)) (x :-> V.+1 :+ (y :-> V :+ h))
   by move=> h1 v1 w1; case/(cancel D)=>-> D2; case/(cancel D2)=>-> _ ->.
 by case: (H _ P D1)=>[P0 H1]; apply: val_do=>// [tt1] {k} k Dk;
   move/H1/(_ Dk); apply: False_ind.
move=> c k [k1][_]; case; move/(_ h); rewrite -!unA.
move/(_ (erefl _))=>-> [->]H; apply: vrfD=>D1.
apply: bnd_do=>//=; first (by hhauto). 
move=> {k k1} tt k; move/(_ ( y :-> V.+1 :+ h) V (erefl _))=>[->]_.
have P: mswapP1 (x :-> V :+ (y :-> V.+1 :+ h)) (x :-> V.+1 :+ (y :-> V.+1 :+ h))
 by  move=> h1 v1 w1; move/(cancel D1)=>[<-]D2; case/(cancel D2)=> <- _ ->.
apply: vrfD; case/( H _ P)=>[P1 H1]; apply: val_do=>//= tt1{k}k Dk; case: tt1.
move/H1/(_ Dk)=>/= [_]-> h1; rewrite -!unA.
by  case/(cancel D)=> _ D2; case/(cancel D2)=> _ _ <-. 
Qed.
End MimicAbort.