(***************************************************)
(*    HTTcc Examples for the ICFP'13 submbission   *)
(*                rember-up-to-last                *)
(***************************************************)

From mathcomp Require Import ssreflect ssrfun ssrbool ssrnat eqtype seq div.
From Kont Require Import rels prelude heaps domains hprop callcc srules.
Require Import Eqdep.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(* We unfold the defininiton of separation star (#), this, \In, etc
and deal with explicit heap equality and heap union :+ :

 p1 # p2 : Pred heap :=
Pred_of_Simpl
  [Pred h | exists h1 h2 : heap, h = h1 :+ h2 /\ h1 \In p1 /\ h2 \In p2]

this h  : Pred heap := Pred_of_Simpl [Pred h | i = h]

i \In P <-> P i

Other relevant definitions and properties can be found in the heaps.v
and hprop.v. For further details on def predicates, see

[26] NANEVSKI, A., VAFEIADIS, V., AND BERDINE, J. Structuring the
verification of heap-manipulating programs. In POPL (2010). *)
Open Scope ccscope.

Parameters (X: lattice) (A : eqType).
(* some lattice return type, and a type A suppoting equality *)

(* Section 6.1. Remeber-up-to-last*)

Section rember.
Variable (x : A).


(* The tail recursion implementation of rember, from Figure~6 *)
Fixpoint rember acc xs :=
  if xs is y :: ys' then 
    if x == y then rember [::] ys' 
    else rember (rcons acc y) ys'
  else acc. 

(* remberP from Figure 5 *)
Fixpoint remberP acc qs : Pred _ := 
  if qs is z :: zs then 
    if x == z then xPred0 else remberP (rcons acc z) zs
  else eq^~ acc.

(* The lemmas about rember and remberP used in the Proof in Section 6.1  *)
Lemma rember_cat acc ps qs : 
        rember acc (ps ++ qs) = rember (rember acc ps) qs. 
Proof. by elim: ps acc=>[|p ps IH] acc //=; case: eqP. Qed.

Lemma rember_inE acc ps :
        rember acc ps = if x \in ps then rember [::] ps else acc ++ ps.
Proof.
elim: ps acc=>[|p ps IH] acc /=; first by rewrite cats0.
rewrite inE /= -cat_rcons; case: eqP=>[->|_] //=.
by rewrite IH (IH [:: p]); case: ifP.
Qed.

Lemma rember_consP y acc ps qs : 
        qs \In remberP acc ps -> 
        y :: qs \In remberP (y :: acc) ps.
Proof.
elim: ps acc=>[|p ps IH] /= acc; first by move=>->. 
by case: eqP=>// _; move/IH.
Qed.

Lemma rember_remberP acc ps qs :
        qs \In remberP acc ps -> qs = rember acc ps.
Proof. by elim: ps acc=>[|p ps IH] acc //=; case: eqP=>// _; move/IH. Qed.

End rember.



(* The type invariant remberT for Figure~5:

remberT := forall ys : list A, 
      SK* {this j _/\_ exists ps, xs = ps ++ ys} (list A)
          { this j _/\_ r \In remberP x ys []}

is translated as (mind the argument swap in remberP)

 *)


Definition rempT j x xs : Type := forall (ys: seq A), 
  SK X (fun i => 
           i = j  /\ exists ps, xs = ps ++ ys, 
              fun rs i m =>   
                   i = j -> forall ps, xs = ps ++ ys ->
                m = i /\ rs \In remberP x [::] ys).

(* The types for the continuation object exit,

exit : [h]. Kont* {j \In this h}
                  { if x \In xs then this h else \bot}
                  (list A) 
                  { this h _/\_ r = rember x [] xs}

and for remebr-up-to-last


remebr-up-to-last x xs : [h] SK {this h} { this h _/\_ 

are simplified after translating them, as the logical variable h is no
longer needed to name the initial heap with binary posts

e.g. exists h, i = h is always true, thus

rember-up-to-last can be executed in any heap

*)
Program
Definition remberuptolast x xs :
             SK X (xPredT, 
                    fun rs i m => m = i  
                                  /\ rs = rember x [::] xs) :=
Do (callcc (fun j i =>  forall h,  j =  h ->
                        if (x \in xs) 
                       then  i = j else False)
       (fun j  rs i m =>  i = j -> m = i 
                         /\ rs = rember x [::] xs)
   (fun j exit => ffix (fun (f: rempT j x xs) qs =>
      Do (CCC.Match_seq qs (ret [::])
           (fun y ys =>
                     zs <-- f ys ;
                     If x == y Then  throw exit zs
                               Else ret (y::zs)))) xs)).
Next Obligation.
move=> i /= [<-][ps]{exit f}->{xs}.
case: qs=>[|y ys]; first (by apply: val_ret); rewrite -cat_rcons.
apply: bnd_do=>//=[|zs k]; first (by split=>//; exists (rcons ps y)).
move/(_ (erefl _)); case/(_ _ (erefl _))=>->{j} H. 
case: eqP H=>[->{x}|N] H; last  first.
+ by apply: val_ret=> * ; split=>//; apply: rember_consP.
apply: val_throw=>//; split=>//=; [ move=> h ->|];
rewrite cat_rcons ; [ rewrite  mem_cat | rewrite rember_cat rember_inE];
rewrite  inE /= eq_refl /=; first by (case: orP=>//; apply; right).
by apply: rember_remberP.
Qed.
Next Obligation.
move=>i _ ; apply: val_ccc.
+ apply: val_do=>//=[|ps k _]; first by (split=>//;exists [::]).
 by case/(_ (erefl _) [::] (erefl _))=>->; split=>//; apply: rember_remberP.
move=>ps m _ [j] [_]; case;  move/(_ i (erefl _)) ;case: ifP=>// _ ->. 
by case/(_ (erefl _))=> ->//->[->]->.
Qed.
