(***************************************************)
(* Examples for the  non algebraic   callc  model  *)
(*   LARGE-FOOTPRINT SEMANTICS VERSION             *)
(*  call make clean; make WH=1  first              *)  
(***************************************************)

From mathcomp Require Import ssreflect ssrfun ssrbool ssrnat eqtype seq. 
From Kont Require Import rels prelude heaps hprop callccWH srulesWH.
Require Import Eqdep.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Open Scope ccscope.

(* Ping / Pong Coroutines *)

(* The motivation from this comes also from Thielecke's thesis where
he shows the "ping / pong" pattern of cooperative threading w/o
storing the continuation in the state. *)

(* We reckon that in the Iteration examples, those which used an
auxiliary variable to keep track of changes in the state involved a
certain notion of interleaving cooperation as y got incremented before
x and so on. Thus, we can exploit that pattern to implement "ping
/pong" corroutines. *)

(* In Haskell using HTTcc monad, i.e. a \hat{S}K monad with algebraic
call/cc, we define the following: *)

(* pongLoop f n  *)
(*          | n == 0    = return $ update (++ "pong <>") >>  return 1 *)
(*          | otherwise = return $  do *)
(*                                  update (++ "pong -> ") *)
(*                                  abort f $ pongLoop f (n -1) *)

(* progPingPongN n = do *)
(*   c <- callcc $ \k -> pongLoop k n  *)
(*   update (++ "ping -> ") *)
(*   c *)
(*   get *)

(* pingPongN m = prog (progPingPongN m) "" (\ x y -> x) *)

(* We replace Ping and Pong for functions that increment the x and y
 variables by one: after n iterations of the continuation/loop
 interaction, both x and y have been incremented by 1 *)


Program Definition incr ( x : ptr) : 
   ST  ( fun i => exists h, exists v: nat , i = x :-> v :+ h , 
          fun y i m => forall h v, i = x :-> v :+ h -> 
                         m = x :-> v .+1 :+ h /\ y = tt) :=
Do ( z <-- ! x ; x ::= z .+1 ).
Next Obligation.
move=> i [h][v]->; apply: bnd_read; apply: vrfD=>D; apply:val_write=> _ h1 v1.
by case/(cancel D)=>->_->; split=>//.
Qed.

Parameters x y : ptr.
Definition ping := incr x.
Definition pong := incr y.

Parameters xV yV : nat.

(* We mimick the behaviour of abort in the algebraic model by binding the *)
(* intermediate computation before throwing *)

Section ThrowPong.

Definition niceL (n : nat) (j :  heap) : Type := 
  forall z: nat , ST 
     ( fun i => exists h, exists v: nat,
          i = x :-> (xV + v .+1 ) :+  y :-> (yV + v) :+ h 
                 /\ j = x :-> xV :+ y :-> yV  :+ h 
                 /\ z <= n
                 /\ v.+1 + z = n .+1,
           fun (w: nat) i m => 
            forall v h,
             j = x :-> xV :+ y :-> yV :+ h ->
              i = x :-> (xV + v .+1 ) :+  y :-> (yV + v) :+ h ->
                 if (z == 0 ) then 
               m = x :-> (xV + v .+1 ) :+  y :-> (yV + v.+1) :+ h
                  /\ w = 0 else False).

Definition niceP  j : pre :=
    fun i => forall h (v: nat),
             j = x :-> (xV + v) :+  y :-> (yV + v) :+ h ->
             i = x :-> (xV + v .+1) :+  y :-> (yV + v) :+ h.

Definition niceQ ( z n: nat) : post nat := 
  fun r (i m : heap) => 
    forall h (v w : nat), 
      i = x :-> v :+ y :-> w :+ h -> 
      if ( z == xV + n .+1 ) then 
        m = x :-> z :+ y :-> (yV + n.+1) :+ h /\ r = 0 else False.
                      
Program
Definition throwPong ( n : nat) : ST ( fun i =>  
    exists h, 
              i = x :-> xV :+ y :-> yV :+ h,
         fun z i m =>  forall h ( v w : nat),
                i = x :-> v :+ y :-> w :+ h ->
                 m = x :-> ( v + n .+1)  :+ y :-> (w + n.+1) :+ h 
                    /\ z = 0) :=
  Do ( c <-- @CCC.callcc _

(* sf *)
       (fun i => exists h, i = x :-> xV :+ y :-> yV :+ h,
         fun r i m =>
                    forall h (v w: nat),
                      i = x :-> v :+ y :-> w :+ h -> 
                      if (n == 0) then 
                        (m = i /\
                        CCC.conseq (spec_of r) (niceP i) (niceQ (xV .+1 ) n)) else
                        ( exists z : nat, 
                          m = x :-> (v + z) :+ y :-> ( w + z) :+ h) /\
                       CCC.conseq (spec_of r) (niceP m) (niceQ (xV + n.+1) n))

              (fun j =>
                (fun i => exists h,
                            i = x :-> xV :+ y :-> yV :+ h /\ j = i,
                 fun z i m => m = i /\
                      CCC.conseq (spec_of z) (niceP j) (niceQ (xV .+1 ) n)))
    
    (fun j ex1 =>
            Do (ret 
                 (BigST 
                    (ffix (fun (f: niceL  n j) (m: nat) => 
                              Do (CCC.Match_nat m 
                                   (pong ;; ret 0)
                                   (fun z =>
                                           pong ;;
                                           throw ex1 (BigST (f z)) _))) n)))) _;
      ping ;;
      w <-- comm c ;
      ret w).
Next Obligation.
rewrite !gh; apply: ghE. 
- by move=> i [h][v][->][->]_ _; exists (v, h); split=>//.
move=>i {ex1} p; rewrite -!unA=>[[_]]->[h][v]; rewrite -!unA=>[[H0]]HH D.
case: (cancel D H0) HH; rewrite -addn1 -[v.+1]addn1; move/addnI. 
rewrite addnC [v + 1]addnC; move/addnI=>->D2; case/(cancel D2)=> _ _ ->[J]. 
(* case on m // we dont neeed the IH*)
case: m => [_| m0] {H0 D D2}.
(* zero branch*)
- rewrite add1n; apply: vrfD=>D; apply: bnd_do=>/=.
 + by exists (x:-> (xV + v.+1) :+ h); exists (yV + v); rewrite unCA.
 case=> k; move/(_ (x :-> (xV + v.+1) :+ h) (yV + v)); rewrite unCA. 
 by case/(_ (erefl _))=>-> _; apply: val_ret=>//; rewrite unCA  -addnS.
(* m = S m' branch, we procced by case analysis on n *)
case: n f  =>{p} [| n0] f [H3 H4]; first by discriminate.
rewrite unCA; apply: bnd_do=>//=. 
- by  exists (x :-> (xV + v.+1) :+ h); exists (yV + v); hhauto.
case=> k; move/( _ ( x :-> (xV + v.+1) :+ h) (yV +v)); case/(_ (erefl _))=>-> _.
rewrite unCA; apply: vrfD=>D1; apply: val_throw=>h1 v1 w1 H5.
rewrite J -unA in H5; split.
- exists v.+1.
 have D2: def (x :-> xV :+ (y :-> yV  :+ h))  
  by rewrite !defPtUn !domPtUn !defPtUn in D1 *.
 by rewrite unA -addnS; case:(cancel D2 H5)=><- D3; case/(cancel D3)=><- _ <-.
rewrite !gh; apply: ghE.
+ move=> i0; move/(_ h v.+1); rewrite !addnS unA; move/( _ (erefl _))=> -> _.
 by exists (h, (xV + v).+2, (yV + v).+1).
move=>{k D1} k p2;rewrite -unA addnS=>H; move/(_ h v.+1); rewrite !addnS -!unA.
move/(_ (erefl _))=>K _; rewrite K in H *;apply: vrfD=>D2;apply: val_do=>/=.
+ exists h; exists v.+1; rewrite  -!unA !addnS; move: (ltnW H3).
 by rewrite  -[m0.+1]addn1 add1n [m0 + 1]addnC addnA addn1 in H4. 
move=>z k1 _;move/(_ v.+1 h); rewrite -!unA -!addnS;move/(_ J (erefl _)).
case:eqP=>//= H6 [-> ->].
rewrite H6 addn1 add1n in H4 *; rewrite H4; case:eqP=>//= _.
by case: (cancel D2 H)=> _ D3; case/(cancel D3)=> _ _ -> {D3}.
Qed.
Next Obligation.
move=>/= i[h][->]J; rewrite -unA; apply: vrfD=>D; apply: val_ret; split=>//.
move=> k1; move/(_ h 0); rewrite !addn0; move/(_ J)=>->.
rewrite addn1; apply: val_do=>//=.
+ by  exists h; exists 0; rewrite addn0 [1 + n]addnC !addn1.
move=> w k2 Dk; move/(_ 0 h J); rewrite addn0 addn1 -!unA; move/(_ (erefl _)).
case: eqP=>//= -> [->]-> h1 v1 w1; rewrite !addn1 -!unA. 
have D1 : def (x :-> xV.+1 :+ (y :-> yV :+ h)) by
    rewrite !defPtUn !domPtUn !defPtUn in D *.
by case/(cancel D1)=> _ D2; case/(cancel D2)=>{D2} _ _ ->; case: eqP=>//.
Qed.
Next Obligation.
rewrite !gh; apply: ghE; first by move=> i [_][h]-> _; exists (h, xV, yV).
move=> i p; rewrite -!unA=> -> [J][h]; rewrite -unA=> H0 D.
case: (cancel D H0) J (D)=>-> D1; case/(cancel D1)=>{D D1} -> _ -> J D.
apply: val_do; first by exists h; split=>//; rewrite unA.
move=>/= c k _ [->]; case: n J=>[|n0] -> H1; case:eqP=>// _.
split; first by exists 0; rewrite !addn0 unA.
have P1: niceP (x :-> xV :+ (y :-> yV :+ h)) (x :-> xV.+1 :+ (y :-> yV :+ h)).
- move=> h1 v; rewrite -!unA; case/(cancel D)=> H D1 H2.
 rewrite -addn1 {1}H addn1 addnS; case: (cancel D1 H2)=>{H D1 H2}H3 _ ->.
 by rewrite {1}H3.
have D1: def (x :-> xV.+1 :+ (y :-> yV :+ h)) by
   by rewrite !defPtUn !domPtUn !defPtUn in D *.
case: (H1 _ P1 D1)=>{k}P H2 k; move/(_ h 0); rewrite !addn0 -!unA addn1. 
move/(_ (erefl _))=>->; apply: val_do=>//{k} m k Dk Hs.
move:(@H2 m k Hs Dk); move/(_ h xV.+1 yV);rewrite -!unA; move/(_ (erefl _)).
rewrite -addn1 -[n0.+2]addn1; case: eqP=>//=. 
(* n.+2 /= 1 *)
by move/addnI; rewrite addnC -{1}[1]addn0 ;move/addnI; discriminate.
Qed.
Next Obligation.
rewrite !gh; apply: ghE; first by move=>i [h]-> _; exists (h, xV, yV).
move => i p ->[h]; rewrite -!unA=>H D;case: (cancel D H)=>->D2.
case/(cancel D2)=>{D D2 H}-> _ ->; apply: vrfD=>D; apply: bnd_do=>//=.
- by exists h; rewrite unA.
(* out of callcc, spec is sf *)
move=> c k; move/(_ h xV yV); rewrite -!unA; move/(_ (erefl _)).
(* If n = 0 there are no jumps *) 
case:eqP=>//[->[->]| N [[z]]->];[ move=>H | rewrite -!unA=> H].  
- apply: bnd_do=>//=; first by exists (y :-> yV :+ h); exists xV.
 case=>{k} k; case/(_ (y :-> yV :+ h) xV (erefl _))=>-> _.
 have P1: niceP (x :-> xV :+ ( y :-> yV :+  h)) 
                   (x :-> xV.+1 :+ ( y :-> yV :+ h))  by
   move=> h2 v2; rewrite -!unA addnS; move/(cancel D)=>[<-]_->.
 apply:vrfD; case/(H _ P1)=>[P H2];apply: bnd_do=>//= {k} m k H3; apply:vrfD. 
 move/(H2 m k H3); move/(_ h xV.+1 yV); rewrite -!unA !addn1;move/(_ (erefl _)).
 by case: eqP=>//==> _ [->]->; apply: val_ret.
(* case n > 0 ; there is at least one iteration *)
apply: vrfD=>{D}D. 
have P1: niceP (x :-> (xV + z) :+ (y :-> (yV + z) :+ h))
                 (x :-> (xV + z).+1 :+ ( y :-> ( yV + z) :+ h)).
+ move=> h1 v1; rewrite addnS -!unA;case/(cancel D)=><- D1; case/(cancel D1).
 by move=> -> _ ->. 
apply: bnd_do=>//=; first (by hhauto); case=>{k} k.
case/(_ (y :-> (yV + z) :+h) (xV + z) (erefl _))=> -> _; apply: vrfD=>D2.
case: (H (x :-> (xV + z).+1 :+ ( y :-> (yV + z) :+ h)) P1 D2)=>[ P2 H3].
apply: bnd_do=>//= w0 k1 H4; apply: vrfD=>Dk;  move: (H3 w0 k1 H4 Dk).
move/(_ h (xV + z).+1 (yV + z)); rewrite -unA; move/(_ (erefl _)).
by case:eqP =>// _ [->->]; apply: val_ret=>//; rewrite unA. 
Qed.

End ThrowPong.

(* Can we simplify the specs like we did for incr3 *)

Section Throwv2.
     
Let sf n := (fun i => exists h, i = x :-> xV :+ y :-> yV :+ h,
         fun r i m =>
            forall h (v w: nat),
             i = x :-> v :+ y :-> w :+ h -> 
              if (n == 0) then 
                (m = i /\
                 CCC.conseq (spec_of r) (niceP i) (niceQ (xV .+1 ) n)) else
                ( exists z : nat, 
                  m = x :-> (v + z) :+ y :-> ( w + z) :+ h) /\
                 CCC.conseq (spec_of r) (niceP m) (niceQ (xV + n.+1) n)).

Program
Definition throwPongv2 ( n : nat) : ST ( fun i =>  
    exists h, 
              i = x :-> xV :+ y :-> yV :+ h,
         fun z i m =>  forall h ( v w : nat),
                i = x :-> v :+ y :-> w :+ h ->
                 m = x :-> ( v + n .+1)  :+ y :-> (w + n.+1) :+ h 
                    /\ z = 0) :=
  Do ( c <-- @CCC.callcc _ (sf n)

              (fun j => (fun i => j = i /\ (sf n).1 i, (sf n).2))   
    (fun j ex1 =>
            Do (ret 
                 (BigST 
                    (ffix (fun (f: niceL  n j) (m: nat) => 
                              Do (CCC.Match_nat m 
                                   (pong ;; ret 0)
                                   (fun z =>
                                           pong ;;
                                           throw ex1 (BigST (f z)) _))) n)))) _;
      ping ;;
      w <-- comm c ;
      ret w).
Next Obligation.
rewrite !gh; apply: ghE. 
- by move=> i [h][v][->][->]_ _; exists (v, h); split=>//.
move=>i {ex1} p; rewrite -!unA=>[[_]]->[h][v]; rewrite -!unA=>[[H0]]HH D.
case: (cancel D H0) HH; rewrite -addn1 -[v.+1]addn1; move/addnI. 
rewrite addnC [v + 1]addnC; move/addnI=>->D2; case/(cancel D2)=> _ _ ->[J]. 
(* case on m // we dont neeed the IH*)
case: m => [_| m0] {H0 D D2}.
(* zero branch*)
- rewrite add1n; apply: vrfD=>D; apply: bnd_do=>/=.
 + by exists (x:-> (xV + v.+1) :+ h); exists (yV + v); rewrite unCA.
 case=> k; move/(_ (x :-> (xV + v.+1) :+ h) (yV + v)); rewrite unCA. 
 by case/(_ (erefl _))=>-> _; apply: val_ret=>//; rewrite unCA  -addnS.
(* m = S m' branch, we procced by case analysis on n *)
case: n f  =>{p} [| n0] f [H3 H4]; first by discriminate.
rewrite unCA; apply: bnd_do=>//=. 
- by  exists (x :-> (xV + v.+1) :+ h); exists (yV + v); hhauto.
case=> k; move/( _ ( x :-> (xV + v.+1) :+ h) (yV +v)); case/(_ (erefl _))=>-> _.
rewrite unCA; apply: vrfD=>D1; apply: val_throw=>h1 v1 w1 H5.
rewrite J -unA in H5; split.
- exists v.+1.
 have D2: def (x :-> xV :+ (y :-> yV  :+ h))  
  by rewrite !defPtUn !domPtUn !defPtUn in D1 *.
 by rewrite unA -addnS; case:(cancel D2 H5)=><- D3; case/(cancel D3)=><- _ <-.
rewrite !gh; apply: ghE.
+ move=> i0; move/(_ h v.+1); rewrite !addnS unA; move/(_ (erefl _))=> -> _.
 by exists (h, (xV + v).+2, (yV + v).+1).
move=>{k D1} k p2;rewrite -unA addnS=>H; move/(_ h v.+1); rewrite !addnS -!unA.
move/(_ (erefl _))=>K _; rewrite K in H *;apply: vrfD=>D2;apply: val_do=>/=.
+ exists h; exists v.+1; rewrite  -!unA !addnS; move: (ltnW H3).
 by rewrite  -[m0.+1]addn1 add1n [m0 + 1]addnC addnA addn1 in H4. 
move=>z k1 _;move/(_ v.+1 h); rewrite -!unA -!addnS;move/(_ J (erefl _)).
case: eqP=>//= H6 [-> ->].
rewrite H6 addn1 add1n in H4 *; rewrite H4; case:eqP=>//= _.
by case: (cancel D2 H)=> _ D3; case/(cancel D3)=> _ _ -> {D3}.
Qed.
Next Obligation.
rewrite !gh; apply: ghE; first by move=> i [_][h]-> _; exists (h, xV, yV).
move=> i p; rewrite -!unA=> -> [J][h]; rewrite -unA=> H0 D.
case: (cancel D H0) J (D)=>-> D1; case/(cancel D1)=>{D D1} -> _ -> J D.
apply: val_ret; case: ifP. 
+ move/eqP=>/=->; split=>// k; move/(_ h 0); rewrite -!unA !addn0.
 move/(_ (erefl _))=>->; apply: val_do=>//=.
 - by exists h; exists 0; rewrite -!unA !addn0.
 move=> {k} n0 k _; move/(_ 0 h); rewrite -!unA addn0.
 case/(_ J (erefl _))=>-> -> h1 v1 w1; rewrite !addn1 -!unA.
 have D1: def  (x :-> xV.+1 :+ (y :-> yV :+ h)) by rewrite !defPtUn in D *.
 by case/(cancel D1)=> _ D2; case/(cancel D2)=>{D1 D2} _ _ <-; case:eqP.
move=>HH.
split; first by exists 0; rewrite !addn0 unA.
move=> k; move/(_ h 0); rewrite -!unA !addn0; move/(_ (erefl _))=>->.
apply: val_do=>/=; first by exists h; exists 0; rewrite -!unA !addn0.
move=> n0 {k} k _; move/(_ 0 h); rewrite -!unA !addn0.
(* n /= 0 *)
move/(_ J (erefl _)); case:eqP=>// H1; move: HH; case/negP; case:eqP=>//.
Qed.
Next Obligation. by []. Qed.
Next Obligation.
rewrite !gh; apply: ghE; first by move=> i [h]-> _; exists (h, xV, yV).
move=> i p; rewrite -!unA=> -> [h]; rewrite -!unA=>H0 D.
case: (cancel D H0) (D)=>-> D1; case/(cancel D1)=>{D D1} -> _ -> D.
apply: bnd_do=>/= ;first by exists h; rewrite unA.
(* out of callcc, spec is (sf n)*)
move=> c k; move/(_ h xV yV); rewrite -!unA; move/(_ (erefl _)).
(* If n = 0 there are no jumps *) 
case: eqP=>//[->[->]| N [[z]]->];[ move=>H | rewrite -!unA=> H].  
- apply: bnd_do=>//=; first by exists (y :-> yV :+ h); exists xV.
 case=>{k} k; case/(_ (y :-> yV :+ h) xV (erefl _))=>-> _.
 have P1: niceP (x :-> xV :+ ( y :-> yV :+  h)) 
                   (x :-> xV.+1 :+ ( y :-> yV :+ h))  by
   move=> h2 v2; rewrite -!unA addnS; move/(cancel D)=>[<-]_->.
 apply:vrfD; case/(H _ P1)=>[P H2];apply: bnd_do=>//= {k} m k H3; apply:vrfD. 
 move/(H2 m k H3); move/(_ h xV.+1 yV); rewrite -!unA !addn1;move/(_ (erefl _)).
 by case: eqP=>//==> _ [->]->; apply: val_ret.
(* case n > 0 ; there is at least one iteration *)
apply: vrfD=>{D}D. 
have P1: niceP (x :-> (xV + z) :+ (y :-> (yV + z) :+ h))
                 (x :-> (xV + z).+1 :+ ( y :-> ( yV + z) :+ h)).
+ move=> h1 v1; rewrite addnS -!unA;case/(cancel D)=><- D1; case/(cancel D1).
 by move=> -> _ ->. 
apply: bnd_do=>//=; first (by hhauto); case=>{k} k.
case/(_ (y :-> (yV + z) :+h) (xV + z) (erefl _))=> -> _; apply: vrfD=>D2.
case: (H (x :-> (xV + z).+1 :+ ( y :-> (yV + z) :+ h)) P1 D2)=>[P2 H3].
apply: bnd_do=>//= w0 k1 H4; apply: vrfD=>Dk;  move: (H3 w0 k1 H4 Dk).
move/(_ h (xV + z).+1 (yV + z)); rewrite -unA; move/(_ (erefl _)).
by case:eqP =>// _ [-> ->]; apply: val_ret=>//; rewrite unA. 
Qed.
End Throwv2.

(*  Can we clean the strcuture out, like we did for the algebraic version ?  *)

(*  No.  We cannot push the ret inside in both branches, and throw is to *)
(* blame here.If we push ret inside, the zero branch of the *)
(* case-expression results: *)

(*  ret (Bigst pong) : ST (sigST unit).   *)

(* Then the (S n) branch has to match that branch. Easy!! one might think *)

(* (fun z => ret ( Bigst (pong ;; throw ex1 ? ) )  : ST (sigST unit) *)

(* ? : SigST unit *)

(* Hum!! Now (f z) doesn't fit the hole, as throw is creating an extra *)
(* monadic layer. In the algebraic case, the outer ret will make the *)
(* abort call behave as a throw, and then (f z) would have the right type. *)

(* We could recursively define the (SigsST unit computation, but that *)
(* would require aoother ffix--- one which would support a Sigst A outer *)
(* layer.  *)


(* Kudos to the algebraic version, then. However, we could still resort *)
(* to the bind then throw hack from Turon et. al. *)

Section Throwv3.

Definition evenP  j : pre :=
    fun i => forall h (v: nat),
             j = x :-> (xV + v) :+  y :-> (yV + v) :+ h ->
             i = x :-> (xV + v .+1) :+  y :-> (yV + v) :+ h.

Definition evenQ ( z n: nat) : post unit := 
  fun r (i m : heap) => 
    forall h (v w : nat), 
      i = x :-> v :+ y :-> w :+ h -> 
      if ( z == xV + n .+1 ) then 
        m = x :-> z :+ y :-> (yV + n .+1) :+ h /\ r = tt else False.

Definition evenL (n : nat) (j :  heap) : Type := 
  forall z: nat , @ST (SigST unit) 
     ( fun i => exists h, exists v: nat,
          i = x :-> (xV + v ) :+  y :-> (yV + v) :+ h 
                 /\ j = x :-> xV :+ (y :-> yV  :+ h )
                 /\ z <= n
                 /\ v + z = n,
           fun r i m => forall (v : nat) h,
             i = x :-> (xV + v) :+ y :-> (yV + v) :+ h -> 
                       m = i  /\
                       CCC.conseq (spec_of r) (evenP m) (evenQ (xV + n.+1) n)).

Let sf n := (fun i => exists h, i = x :-> xV :+ ( y :-> yV :+ h),
         fun r i m =>
            forall h (v w: nat),
             i = x :-> v :+ (y :-> w :+ h )-> 
              if (n == 0) then 
                (m = i /\
                 CCC.conseq (spec_of r) (evenP i) (evenQ (xV .+1 ) n)) else
                ( exists z : nat, 
                  m = x :-> (v + z) :+ y :-> ( w + z) :+ h) /\
                 CCC.conseq (spec_of r) (evenP m) (evenQ (xV + n.+1) n)).
 
Program
Definition pingPongv3 ( n : nat) : ST ( fun i =>  
    exists h, 
              i = x :-> xV :+ y :-> yV :+ h,
         fun z i m =>  forall h ( v w : nat),
                i = x :-> v :+ y :-> w :+ h ->
                 m = x :-> ( v + n .+1)  :+ y :-> (w + n .+1) :+ h 
                    /\ z = tt) :=
  Do ( c <-- callcc  (sf n)
                 (* (fun j => (fun i => j = i /\ (sf n).1 i, (sf n).2))    *)
            (fun j ex1 =>
                 ffix (fun (f: evenL  n j) (m: nat) => 
                   Do (CCC.Match_nat m 
                       (ret (BigST pong))
                       (fun z => ret 
                          (BigST (pong ;; 
                                  t <-- (f z);
                                  throw ex1 t _))))) n) _;
      ping ;;
      comm c).
Next Obligation.
rewrite !gh; apply: ghE=>/=.
+ by move=> i [h][v][->] _ _; exists (v, h).
move=> i p -> [h][v][H0][J][H1 H2]; rewrite -!unA in H0 *=>D0.
move: (cancel D0 H0);case; move/addnI=>-> D1; case/(cancel D1)=>{D0 D1} _ _ ->.
rewrite unCA; apply: vrfD => D; case: m H1 H2.
(* We case on m (we don't need the IH *)
(* m = 0 , thus we are in the  (ret (BigST (pong))) branch *)
+ rewrite addn0=> _->;  apply: val_ret; split=>// k; rewrite unCA unA.
 move/(_ h n (erefl _))=>->; apply: val_do=>//=. 
 - by exists ( x :-> (xV + n.+1) :+ h); exists (yV + n); rewrite unCA unA.
 move=> tt {k ex1 f J p H0} k1 _; rewrite -unA unCA.
 move/(_ (x :-> (xV + n.+1) :+ h) (yV + n) (erefl _))=>[K]_; rewrite K in D *.
 move=> h2 v2 w2; rewrite -unA [ x :-> v2 :+ _ ]unCA.
 case/(cancel _); first (by rewrite !defPtUn !domPtUn !defPtUn in D * ).
 move=>_ D0;case/(cancel D0)=>{D0}_ _ <-.
 by rewrite !addnS unCA unA; case: eqP=>// _; split=>//; elim: tt.
(* m = S m0 then, we are going to jump out eventually *)
move=> m0 H1 {p H0}; rewrite addnS=>H2; apply: val_ret; split=>//.
(* the body just returns, but we need to prove the conseq for the returnee. *)
(* This is the first layer of Big, namely abort *)
rewrite !gh; apply: ghE=>/=.
+ move=> k; rewrite unCA unA; move/(_ h v (erefl _))=>-> _.
 by  exists (h, xV + v.+1, yV + v).
move=> k p ->; rewrite unCA unA ; move/(_ h v (erefl _)); rewrite -!unA =>H D1.
move: (cancel D1 H)=>{D1}{H}[->]D2; case/(cancel D2)=>{D2}->_->.
(* we've went somehere and it nows the computation gets evaluated *)
(* pong *)
apply: bnd_do=>//=; first by rewrite unCA;hhauto.
case=> {k}k; move/( _ (x :-> (xV+ v.+1) :+ h) (yV + v)); rewrite unCA.
case/(_ (erefl _))=> -> _; rewrite unCA; apply: bnd_do=>//= {k}.
+ by exists h; exists v.+1; rewrite unA J ltnW// -H2 !addSn -addnS.
move=>c k; move/(_ v.+1 h); rewrite -unA !addnS ;move/(_ (erefl _))=>[->]=> H4.
rewrite unCA; apply: val_throw=> h1 v1 w1; rewrite J.
have D1: def (x :-> xV :+ (y :-> yV :+ h)) by
             rewrite unCA !defPtUn !domPtUn !defPtUn in D *.
case/(cancel D1)=> <- D2; case/(cancel D2)=> <- _ <- {D1 D2 f ex1}.
(* n cannot be zero *)
case: ifP=>[|_]; first by rewrite -H2 -addn1 addn_eq0; case/andP; discriminate.
split; first by exists v.+1; rewrite !addnS unCA unA.
move=> {k} k; move/(_ h (v.+1)).
rewrite unCA -unA !addnS;move/(_ (erefl _))=>->.
have P1: evenP (x :-> (xV + v).+1 :+ (y :-> (yV + v).+1 :+ h))
                 (x :-> (xV + v).+2 :+ ( y :-> ( yV + v).+1 :+ h)).
+ move=>h2 v2; rewrite -!unA.
 have D1: def (x :-> (xV + v).+1 :+ (y :-> (yV + v).+1 :+ h)) by
    rewrite unCA !defPtUn !domPtUn !defPtUn in D *.
 by case/(cancel D1)=>-> D2; case/(cancel D2)=> -> _ ->; rewrite addnS.
rewrite -unA;apply:vrfD;case/(H4 _ P1)=> [H5 H6].
by apply: val_do=>//= [[]] {k}k Dk; move/H6;move/( _ Dk).
Qed.
(* We can't make the second obligation go away automatically as the loop *)
(* invariant has extra information regarding the iterations and n *)
Next Obligation.
rewrite !gh; apply: ghE; first by  move=> i [_][h]-> _; exists (h,xV,yV).
move=> i p -> [->][h]I; rewrite I=>D; apply: val_do=>//=.
- by exists h; exists 0; rewrite unA !addn0 add0n .
move=> c k _; move/(_ 0 h); rewrite -unA !addn0; case/(_ (erefl _))=>->.
case: n =>[|n0];[rewrite addn1|];move=> H; case:eqP=>// _;split=>//.
exists 0; rewrite !addn0 unA.
by case:(cancel D (esym I))=><- D1; case/(cancel D1)=><- _ <-.
Qed.
Next Obligation.
rewrite !gh; apply: ghE; first by move=> i [h]-> _; exists (h,xV,yV).
move => i p -> [h]; rewrite -!unA=> H; rewrite H=>D.
case:(cancel D (esym H))=><- D1; case/(cancel D1)=>{D1}<- _<-.
 apply: bnd_do=>//= {H p}; first by hhauto.
(* sf will be the "loop invariant here"*)
 move=> c k ;move/(_ h xV yV); rewrite unA; move/(_ (erefl _)).
rewrite -!unA;case: n=> [|n0]; case:eqP=>// _; [case=>[->] H|case=> [[z]]->].
(* n = 0*)
+ apply: bnd_do=>/=; first (by hhauto); case=>{k} k.
 (* ping *)
  have P1: evenP (x :-> xV :+  (y :-> yV :+  h))
                   (x :-> xV.+1 :+ (y :-> yV :+ h)).
  - by move=> h2 v2; rewrite -!unA addnS; move/(cancel D)=>[<-]_->.
 move/(_ (y :-> yV :+ h) xV (erefl _))=>[->]_.
 apply: vrfD; case/(H _ P1)=>H1 H2.
 (* closure *)
 apply: val_do=>// {k} [[]] k Dk; move/(H2 _ _ ); move/(_ Dk).
 move/(_ h ( xV .+1) yV); rewrite -unA; move/(_ (erefl _)).
 by case: eqP=>//= _ [->] _; rewrite -unA !addn1.
(* n > 0 *)
rewrite -!unA=> H; apply: bnd_do=>//=; first by hhauto.
(* z + 1 ping *)
case=> {k} k; case/(_ (y :-> (yV + z) :+ h) (xV + z) (erefl _))=> -> _. 
have P1: evenP (x :-> (xV + z) :+ (y :-> (yV + z) :+ h))
                 (x :-> (xV + z).+1 :+ ( y :-> ( yV + z) :+ h)).
- move=>h1 v1; rewrite -!unA addnS. 
 have D1: def (x :-> (xV + z) :+ (y :-> (yV + z) :+ h)) 
    by rewrite !defPtUn !domPtUn !defPtUn in D*.
 by case/(cancel D1)=><- _ <-.
 apply: vrfD; case/(H _ P1)=>H1 H2.
 (* last closure run *) 
 apply:val_do=>// {k} [[]] k Dk; move/(H2 _ _ ); move/(_ Dk).
 move/(_ h ( xV + z.+1) (yV + z)); rewrite -unA addnS; move/(_ (erefl _)).
 by case: eqP=>//= _ [->] _; rewrite -unA.
Qed.
End Throwv3.
